class  UnModbusRedirectEntry {

    public string plcIP;   // CFG_IP
    public int    port ;   // CFG_PORT
    public string project; // CFG_PROJ
    public int    driver;  // CFG_DRIVER
    public string comment; // CFG_COMMENT
    public bool   conf_inFile;// CFG_CONFRED
    public bool   conf_live;// CFG_CONFMOD
    public string hostname;// CFG_HOSTNAME

    public UnModbusRedirectEntry(
	string newPlcIP="", int newPort=0, string newProject="", int newDriver=0,
	string newComment="", bool   newConfInFile=false, bool   newConfLive=false, string newHostname="") 
    {
		plcIP    = newPlcIP;
		port     = newPort;
		project  = newProject;
		driver   = newDriver;
		comment  = newComment;
		conf_inFile = newConfInFile;
		conf_live = newConfLive;
		hostname = newHostname;
    }

    public string serialize() 
    {	
    	return jsonEncode(this,true); // 2nd param means compact version
    }

    public static UnModbusRedirectEntry deserialize(string s)
    {
		mapping m=jsonDecode(s);
		dyn_errClass err=getLastError();
		if (dynlen(err)) {
			throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,"Problem with decoding data entry: "+s));
    		UnModbusRedirectEntry e=UnModbusRedirectEntry();
    		e.comment="DECODING ERROR:"+s;
    		e.plcIP="#";
    		return e;
		}
    	UnModbusRedirectEntry e=UnModbusRedirectEntry(m["plcIP"],m["port"],m["project"],m["driver"],m["comment"],m["conf_inFile"],m["conf_live"],m["hostname"]);
    	return e;
    }
};

class UnModbusRedirectConfiguration {

	public string error="";
	protected dyn_mixed entries;

    public UnModbusRedirectConfiguration()
    {
    }

	public static UnModbusRedirectConfiguration fromFile(string cfgFile)
	{
		UnModbusRedirectConfiguration cfg;
		dynClear(cfg.entries);
		
		if (cfgFile=="") return cfg;
		
		string s;
		bool ok=fileToString(cfgFile,s);
		if (!ok) {
			cfg.error="ERROR while reading the file:"+cfgFile;
			return cfg;
		}
		
		dyn_string ds=strsplit(s,"\n");
		for (int i=1;i<=dynlen(ds);i++) {
			if (ds[i]=="") continue;//empty line
			if (substr(ds[i],0,1)=="#") {
				// a comment; keep a special entry that has "#" as PLC IP
				string comment=strltrim(ds[i],"# ");
				UnModbusRedirectEntry e=UnModbusRedirectEntry("#",0,"",0,comment,true,false,getHostname());
				dynAppend(cfg.entries,e);
				continue;
			}

			dyn_string tokens;
			//int rc=regexpSplit("^\\s*(\\S+)\\s+(\\S+)\\s+(\\d+)\\s+(\\S+)\\s+(\\#*)\\s*(.+)",ds[i],tokens);
			int rc=regexpSplit("^\\s*(\\S+)\\s+(\\S+)\\s*(\\S*)\\s*(\\S*)\\s*(#*)(\\s*)(.+)?",ds[i],tokens);
			
			// Note that tokens[1] matches the complete string, hence we get captured 
			// substring starting at index 2
			//
			// The regexp above works as follows
			// ^ start of the string
			// \s* any amount of white space, or no white space
			// (\S+) PLC Address string => captured into tokens[2]
			// \s+   whitespace (at least one)
			// (\S+) Port number => captured into tokens[3]
			// \s*   whitespace (optional)
			// (\S*) Project name (optional) => captured into tokens[4]
			// \s*   whitespace (optional)
			// (\S*) driver number (optional) => captured into tokens[5]
			// \s*   whitespace (optional)
			// (#*)  #-separator of the comment (optional) => captured into tokens[6]
			// \s*   whitespace that optionally follows the "#" => captured into tokens[7]
			// (.+)? comment (optional), any number of words (also none) => captured into tokens[8]
			
			if (rc == -1) {
				// could not parse the file; $1 is line number, $2 is the entry
			  	throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2001,(string)i ,ds[i]));
			  	continue;
			}
			
			UnModbusRedirectEntry e=UnModbusRedirectEntry(
				tokens[2],      //plc address
				(int)tokens[3], //port number
				tokens[4],      //project
				(int)tokens[5], //driver number
				tokens[8],      //comment
				true,			//conf_inFile
				false,			//conf_live
				getHostname()	//host
				);
			dynAppend(cfg.entries,e);
		}
		
		cfg.error="";
		return cfg;
	}

	public int numEntries()
	{
		return dynlen(entries);
	}
	
	/**
		Note: zero-based indexing!
	*/
	public UnModbusRedirectEntry at(unsigned i)
	{
		UnModbusRedirectEntry e=entries[i+1];
		return e;
	}


	public void clear()
	{
		dynClear(entries);
	}
	
	public void addEntry(UnModbusRedirectEntry e)
	{
		dynAppend(entries,e);
	}

    public dyn_mixed getConfiguration() {
    	if (error!="") return makeDynString();
		return entries;
    }

    public void saveToFile(string cfgFile)
    {
    	string newContent;

    	bool ok;
    	// check if file exists, make a backup, preserve headers
    	int rc=access(cfgFile,R_OK);
    	if (rc) {
    		// file does not exist... create one from template
			string templCfgFile=getPath(DATA_REL_PATH,"unModbusRedirect/redirectModbus.cfg.template");
        	if (templCfgFile=="") {error="TEMPL_CFG_FILE_DOES_NOT_EXIST";return;}
        	ok=copyFile(templCfgFile,cfgFile);
        	if (!ok) {error="COULD_NOT_CREATE_CFG_FILE_FROM_TEMPLATE";return;}
        }

        ok=copyFile(cfgFile, cfgFile+".bak");
        if (!ok) {error="BACKUP OF CONFIG FILE FAILED";return;}
        
        string s;
        bool ok=fileToString(cfgFile,s);
        if (!ok) {error="Could not open config file for reading";return;}
        

		// and then iterate on what we have
		for (int i=1;i<=dynlen(entries);i++) {
			UnModbusRedirectEntry e=entries[i];
			if (i!=1) newContent+="\n"; // end the previous line
			// special treatment for comment lines:
			if (e.plcIP=="#") {
				newContent+="# "+e.comment;
				continue;
			}
			newContent+=e.plcIP;
			newContent+="\t"+e.port;
			if (e.project=="") continue;
			newContent+="\t"+e.project;
			newContent+="\t"+e.driver;
			if (e.comment=="") continue;
			newContent+="\t# "+e.comment;
		}
		if (dynlen(entries)>0) newContent+="\n"; // append newline at the last line, if there are any entries
		file f=fopen(cfgFile,"w");
		if (f==0) {error="Could not open cfg file for writing";return;}
		int rc=fputs(newContent,f);
		if (rc) error="Could not write to config file";
		fclose(f);
    }

	public string serialize()
	{
		string s;
		for (int i=1;i<=dynlen(entries);i++) {
			if (i!=1) s+="\n";
			s+=entries[i].serialize();
		}
		return s;
	}

	public static UnModbusRedirectConfiguration deserialize(string data)
	{
		UnModbusRedirectConfiguration cfg;
	
		dynClear(cfg.entries);

		dyn_string ds=strsplit(data,"\n");
		for (int i=1;i<=dynlen(ds);i++) {
			UnModbusRedirectEntry e=UnModbusRedirectEntry::deserialize(ds[i]);
			dynAppend(cfg.entries,e);
		}
		cfg.error="";
		return cfg;			
	}

	public int find(string plcip, int port)
	{
		// iteration using 0-based at()
		for (int i=0;i<dynlen(entries);i++) {
			UnModbusRedirectEntry e=at(i);
			if ( (e.plcIP==plcip) && (e.port==port) ) return i;
		}
		return -1; // not found
	}

};
